﻿using RabbitMQ.Client;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Configuration;
using RabbitMQ.Client.Events;
using System.Threading;
using RabbitMQ.Client.Exceptions;
using System.Runtime.InteropServices;
using System.Reflection;
using Utility.CommandLine;

namespace Console_RabbitMQ
{
    class Program
    {
        private Random random;
        private ConnectionFactory factory;
        private IConnection conn;
        private IModel sendChannel;
        private List<string> hostName;
        private List<string> queueList;
        private Action<string, string, int> actionReCreateQueue;
        private CancellationTokenSource cts;
        private Task task;

        // Visit https://github.com/jpdillingham/Utility.CommandLine.Arguments to view more details
        // Usage: cmd.exe -h
		// Usage: cmd.exe -n "10.14.2.51;10.4.1.34;10.4.1.33" -P <server-port> -u <username> -p <password>  -v <virtual-host> -c <Queue-Numbers> -q <queue-name> -t <Quorum/Classic> -m <Consumer/Producer>
        [Argument('n', "hostname", "Host Names or IPs")]
        private static string _hostName { get; set; }

        [Argument('P', "Port", "Server Port")]
        private static int _port { get; set; }

        [Argument('u', "username", "Username")]
        private static string _userName { get; set; }

        [Argument('p', "password", "Password")]
        private static string _passWord { get; set; }

        [Argument('v', "virtualhost", "Virtualhost")]
        private static string _virtualHost { get; set; }

        [Argument('c', "queueCNT", "Queue Counts")]
        private static int _queueCNT { get; set; }

        [Argument('q', "queueName", "Queue Name")]
        private static string _queueName { get; set; }

        [Argument('t', "queueType", "Queue Type")]
        private static string _queueType { get; set; }

        [Argument('m', "mqType", "MQ Type")]
        private static string _rabbitMQType { get; set; }

        [Argument('d', "delay", "Delay Seconds(Default:5)")]
        private static int _delay { get; set; }

        [Argument('s', "size", "Message Size-KB(Default:1)")]
        private static int _messageSize { get; set; }

        [Argument('h', "help", "Show the help.")]
        private static bool Help { get; set; }


        static void Main(string[] args)
        {
            Arguments.Populate();

            _delay = (_delay <1) ? 5 : _delay;
            _messageSize = (_messageSize <1) ? 1 : _messageSize;
            if (Help)
            {
                ShowHelp();

                // It guarantees that you will not proceed because you asked to show help.
                return;
            }
            Console.WriteLine("==================================");
            Console.WriteLine("Hostname: " + _hostName);
            Console.WriteLine("Port: " + _port);
            Console.WriteLine("Username: " + _userName);
            Console.WriteLine("Password: " + _passWord);
            Console.WriteLine("Virtualhost: " + _virtualHost);
            Console.WriteLine("QueueCNT: " + _queueCNT);
            Console.WriteLine("Queuename: " + _queueName);
            Console.WriteLine("Queuetype: " + _queueType);
            Console.WriteLine("RabbitMQType: " + _rabbitMQType);
            Console.WriteLine("Delay: " + _delay);
            Console.WriteLine("MessageSize: " + _messageSize);
            Console.WriteLine("==================================");

            Program prog = new Program();

            prog.random = new Random();
            prog.actionReCreateQueue = prog.ReCreateQueue;

            try
            {


                Console.WriteLine("Connecting...");

                var hostName = _hostName.Split(';').ToList();
                var port = Convert.ToInt16(_port);
                var userName = _userName.ToString();
                var passWord = _passWord.ToString();
                var virtualHost = _virtualHost.ToString();

                prog.CreateConnection(hostName, port, userName, passWord, virtualHost);

                Console.WriteLine("Connect Success");

                var queueCNT = Convert.ToInt16(_queueCNT);
                var queueName = _queueName.ToString();
                var rabbitMQType = (RabbitMQType)Enum.Parse(typeof(RabbitMQType), _rabbitMQType.ToString());
                var queueType = (QueueType)Enum.Parse(typeof(QueueType), _queueType.ToString()); 

                prog.CreateQueue(queueCNT, queueName, rabbitMQType, queueType);

                Console.WriteLine("Create Queue Success");

                if (rabbitMQType == RabbitMQType.Producer)
                {
                    var messageSize = Convert.ToInt16(_messageSize);
                    double delay = 1000 * Convert.ToDouble(_delay);

                    prog.cts = new CancellationTokenSource();

                    prog.task = new Task(() => prog.SendMessage(messageSize, Convert.ToInt16(delay), prog.cts), prog.cts.Token);

                    prog.task.Start();

                }

                Console.WriteLine("輸入任意鍵結束程式");
                Console.ReadKey();

                if (rabbitMQType == RabbitMQType.Producer)
                {
                    foreach (var item in prog.queueList)
                    {
                        prog.DeleteQueue(item);
                    }
                }

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                Console.ReadLine();
            }

        }

        private static void ShowHelp()
        {
            var helpAttributes = Arguments.GetArgumentInfo(typeof(Program));

            Console.WriteLine("Short\tLong\tFunction");
            Console.WriteLine("-----\t------\t--------");

            foreach (var item in helpAttributes)
            {
                var result = "-" + item.ShortName + "\t" + "--" + item.LongName + "\t" + item.HelpText;
                Console.WriteLine(result);
            }
        }

        public void CreateConnection(List<string> hostName, int port, string userName, string passWord, string virtualHost = "/")
        {
            try
            {
                this.hostName = hostName;

                factory = new ConnectionFactory
                {
                    Port = port,
                    UserName = userName,
                    Password = passWord,
                    VirtualHost = virtualHost
                };

                factory.AutomaticRecoveryEnabled = true;

                this.conn = factory.CreateConnection(this.hostName);
                this.conn.ConnectionShutdown += conn_ConnectionShutdown;
                this.sendChannel = this.conn.CreateModel();

            }
            catch (Exception)
            {

                throw;
            }
        }

        public void CreateQueue(int queueCNT, string queueName, RabbitMQType rabbitMQType, QueueType queueType)
        {
            Dictionary<string, object> arg = new Dictionary<string, object>();

            IModel receivedChannel = conn.CreateModel();

            this.queueList = new List<string>();

            switch (queueType)
            {
                case QueueType.Classic:

                    arg.Add("x-queue-leader-locator", "balanced");

                    for (int i = 0; i < queueCNT; i++)
                    {
                        EventingBasicConsumer consumer;

                        var qn = string.Format("{0}{1}", queueName, i);

                        receivedChannel.QueueDeclare(qn, false, false, true, arg);


                        if (rabbitMQType == RabbitMQType.Consumer)
                        {
                            consumer = new EventingBasicConsumer(receivedChannel);

                            receivedChannel.BasicConsume(qn, true, consumer);

                            new Received(
                                consumer,
                                qn,
                                queueType.ToString(),
                                this.actionReCreateQueue);
                        }

                        this.queueList.Add(qn);
                    }


                    break;
                case QueueType.Quorum:

                    arg.Add("x-queue-type", "quorum");
                    arg.Add("x-queue-leader-locator", "balanced");

                    for (int i = 0; i < queueCNT; i++)
                    {
                        EventingBasicConsumer consumer;

                        var qn = string.Format("{0}{1}", queueName, i);

                        receivedChannel.QueueDeclare(qn, true, false, false, arg);

                        if (rabbitMQType == RabbitMQType.Consumer)
                        {
                            consumer = new EventingBasicConsumer(receivedChannel);

                            receivedChannel.BasicConsume(qn, true, consumer);

                            new Received(
                                consumer,
                                qn,
                                queueType.ToString(),
                                this.actionReCreateQueue);
                        }

                        this.queueList.Add(qn);
                    }

                    break;
                default:
                    break;
            }

            if (rabbitMQType == RabbitMQType.Producer)
            {
                receivedChannel.Close();
            }
            else
            {
                sendChannel.Close();
            }

        }

        private void conn_ConnectionShutdown(object sender, ShutdownEventArgs e)
        {
            Console.WriteLine("Connection Shutdown");
        }

        private string GetRandomStr(int length)
        {
            string tmpStr = string.Empty;
            const string characters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

            for (int i = 0; i < Math.Floor(length / Math.Pow(2, 10)); i++)
            {
                tmpStr += new string(Enumerable.Repeat(characters, Convert.ToInt32(Math.Pow(2, 10)))
              .Select(s => s[random.Next(s.Length)]).ToArray());
            }

            return tmpStr;
        }

        private void DeleteQueue(string queueName)
        {
            this.sendChannel.QueueDelete(queueName);
        }

        public void SendMessage(int messageSize, int delay, CancellationTokenSource cts)
        {
            string messageBody = string.Empty;

            try
            {
                using (cts.Token.Register(Thread.CurrentThread.Abort))
                {
                    while (true)
                    {
                        foreach (var item in this.queueList)
                        {
                            messageBody = GetRandomStr(Convert.ToInt16(messageSize * Math.Pow(2, 10)));

                            OnSendMessage(messageBody, "", item, 3);

                        }

                        Task.Delay(delay).Wait();
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("Stop Send Message");
            }

        }

        public void OnSendMessage(string msg, string exchangeName, string routingKey, int retryCNT = 3)
        {
            try
            {
                if (retryCNT > 0)
                {

                    this.sendChannel.BasicPublish(exchangeName, routingKey, null, Encoding.UTF8.GetBytes(msg));

                    Console.WriteLine(string.Format("[{0}]SendMessage : {1}...",
                        routingKey,
                        msg.Substring(0, 8)));

                }
                else
                {
                    new Exception("Send message over retry");
                }


            }
            catch (RabbitMQClientException)
            {
                Console.WriteLine("Retry Send");
                ReConnectionBySender();
                OnSendMessage(msg, exchangeName, routingKey, retryCNT--);
            }
            catch (Exception)
            {
                throw;
            }
        }

        public void ReConnectionBySender()
        {
            this.conn = factory.CreateConnection(this.hostName);
            this.sendChannel.Close();
            this.sendChannel = this.conn.CreateModel();
        }

        public void ReConnectionByReceived()
        {
            this.conn = factory.CreateConnection(this.hostName);
        }

        private void ReCreateQueue(string queueType, string queueName, int retry = 3)
        {
            Console.WriteLine(string.Format("ReCreateQueue [{0}], QueueType [{1}]", queueName, queueType));

            EventingBasicConsumer consumer;

            try
            {

                if (retry > 0)
                {

                    ReConnectionByReceived();

                    IModel receivedChannel = conn.CreateModel(); 

                    Dictionary<string, object> arg = new Dictionary<string, object>();

                    switch (queueType)
                    {
                        case "Classic":

                            arg.Add("x-queue-leader-locator", "balanced");

                            receivedChannel.QueueDeclare(queueName, false, false, true, arg);
                            consumer = new EventingBasicConsumer(receivedChannel);
                            receivedChannel.BasicConsume(queueName, true, consumer);


                            new Received(
                                consumer,
                                queueName,
                                queueType.ToString(),
                                this.actionReCreateQueue
                                );

                            break;

                        case "Quorum":


                            arg.Add("x-queue-type", "quorum");
                            arg.Add("x-queue-leader-locator", "balanced");

                            receivedChannel.QueueDeclare(queueName, true, false, false, arg);

                            consumer = new EventingBasicConsumer(receivedChannel);

                            receivedChannel.BasicConsume(queueName, true, consumer);

                            new Received(
                                consumer,
                                queueName,
                                queueType.ToString(),
                                this.actionReCreateQueue
                                );

                            break;
                        default:
                            break;
                    }
                }
                else
                {
                    throw new Exception("Over Retry");
                }

            }
            catch (Exception ex)
            {

                if (ex.Message.Contains("作業逾時"))
                {

                    Console.WriteLine(string.Format("Retry : {0}", retry));

                    ReCreateQueue(queueType, queueName, retry--);
                }
                else
                {
                    Console.WriteLine(ex.Message);
                }
            }
        }

        public enum QueueType
        {
            Classic,
            Quorum
        }

        public enum RabbitMQType
        {
            Consumer,
            Producer
        }
    }

    public class Received
    {
        private string Name;
        private Action<string, string, int> actionReCreateQueue;
        private string queueType;
        private EventingBasicConsumer consumer;

        public Received(EventingBasicConsumer consumer, string name, string queueType, Action<string, string, int> actionReCreateQueue)
        {
            this.Name = name;
            this.queueType = queueType;
            this.actionReCreateQueue = actionReCreateQueue;
            this.consumer = consumer;

            this.consumer.Received += consumer_Received;
            this.consumer.Shutdown += consumer_Shutdown;
        }

        void consumer_Shutdown(object sender, RabbitMQ.Client.ShutdownEventArgs e)
        {
            Console.WriteLine(string.Format("Consumer[{0}] Shutdown", this.Name));
            this.consumer.Model.Close();
            this.actionReCreateQueue(this.queueType, this.Name, 3);
        }

        void consumer_Received(object sender, BasicDeliverEventArgs e)
        {
            var body = e.Body.ToArray();
            var message = Encoding.UTF8.GetString(body);

            Console.WriteLine(string.Format(" [{0}] Received {1}...",
                this.Name,
                message.Substring(0, 8)));
        }

    }
}
